package projects.CASTOR.nodes.nodeImplementations;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import projects.CASTOR.behaviour.CastorAckBehaviour;
import projects.CASTOR.behaviour.CastorMessageBehaviour;
import projects.CASTOR.messages.CastorAckMessage;
import projects.CASTOR.messages.CastorMessage;
import projects.CASTOR.timers.CastorExpirationTimer;
import projects.CASTOR.timers.CastorSenderTimer;
import projects.common.RoutingBehaviours.RoutingBehaviour;
import projects.common.behaviours.TTLedBehaviour;
import projects.common.nodes.BasicNode;
import projects.common.selector.AlreadySeenSelector;
import projects.common.selector.AndSelector;
import projects.common.selector.TypeSelector;
import sinalgo.nodes.edges.Edge;

/**
 * CASTOR as understood from their paper We do not emulate flows : each node has
 * ONE flow, directed for any sink, which is never exhausted We remember every
 * flow that we see, no expiration date
 * 
 * @author rjamet
 * 
 */
public class CastorNode extends BasicNode {

	/* Parameters from the paper */

	public static double defaultReputationThreshold = .65;
	public static double defaultReputationDecay = .80;
	public static double defaultBroadcastModifier = .65;
	
	public static double defaultTimerLength = 3; //TODO

	public double myReputationThreshold = defaultReputationThreshold;
	public double myReputationDecay = defaultReputationDecay;
	public double myBroadcastModifier = defaultBroadcastModifier;

	public Map<Edge, Map<Integer, ReliabilityEstimator>> reputations = new HashMap<Edge, Map<Integer, ReliabilityEstimator>>();
	public Map<Edge, Set<Integer>> ackReputationMap =  new HashMap<Edge, Set<Integer>>();
	public Map<Integer, BasicNode> ackRoutingMap = new TreeMap<Integer, BasicNode>();
	
	
	@Override
	public void init() {
		super.init();
		
		
		rb = new RoutingBehaviour(this)
				.add(
						new TypeSelector(CastorAckMessage.class),
						new CastorAckBehaviour()
				).add(
						new AndSelector(
								new TypeSelector(CastorMessage.class),
								new AlreadySeenSelector()
							),
							new TTLedBehaviour()
				).add(
						new TypeSelector(CastorMessage.class),
						new CastorMessageBehaviour()
				);
		
		
	}
	
	@Override
	public void neighborhoodChange() {
		super.neighborhoodChange();
		initializeFlows();
		
	}
	@Override
	protected void startTimers() {
		if (!isSink()) {
			(new CastorSenderTimer(this)).startRelative(1, this);
		}
	}

	
	/*
	 * Reputation counter as in the paper, call updateReputation to manipulate
	 * and getValue to use
	 */
	public class ReliabilityEstimator {
		double ackDidNotReturn = 1;
		double ackDidReturn = 0;
		double ackDidNotReturnFirst = 1;
		double ackDidReturnFirst = 0;

		public void updateReputation(boolean didReturn, boolean didReturnFirst) {
			assert (!(didReturnFirst && !didReturn)) : "Ack did either not return OR return first";
			if (didReturn) {
				if (didReturnFirst) {
					ackDidReturnFirst = ackDidReturnFirst * myReputationDecay
							+ 1;
					ackDidNotReturnFirst = ackDidNotReturnFirst
							* myReputationDecay;
				} else {
					ackDidReturnFirst = ackDidReturnFirst * myReputationDecay;
					ackDidNotReturnFirst = ackDidNotReturnFirst
							* myReputationDecay + 1;
				}
				ackDidReturn = ackDidReturn * myReputationDecay + 1;
				ackDidNotReturn = ackDidNotReturn * myReputationDecay;
			} else {
				ackDidReturn = ackDidReturn * myReputationDecay;
				ackDidNotReturn = ackDidNotReturn * myReputationDecay + 1;
				ackDidReturnFirst = ackDidReturnFirst * myReputationDecay;
				ackDidNotReturnFirst = ackDidNotReturnFirst * myReputationDecay
						+ 1;
			}
		}

		public double getValue() {
			double v = 0;
			v += (ackDidReturn / (ackDidReturn + ackDidNotReturn));
			v += (ackDidReturnFirst / (ackDidReturnFirst + ackDidNotReturnFirst));
			v /= 2;
			assert (0 <= v && v <= 1) : "Reputation out of [0..1] ?";
			return v;
		}

	}

	
	public void reputationIncrease(Edge e, int flowID) {
		Double r = reputations.get(e).get(flowID).getValue();
		r = r * myReputationDecay + (1 - myReputationDecay);
	}

	public void initializeFlows() {
		assert(reputations.isEmpty()):"Double initialization of flows";
		for (Edge e : getConnections()) {
			reputations.put(e, new TreeMap<Integer, ReliabilityEstimator>());
			ackReputationMap.put(e, new TreeSet<Integer>());
			for (int flowID = -5; flowID < 1000; flowID++) {
				reputations.get(e).put(flowID, new ReliabilityEstimator());
			}
		}
	}

	public double getMaxReputation(int flowID) {
		double pmax = -1;
		for (Edge e : getConnections()) {
			double p = reputations.get(e).get(flowID).getValue();
			if (p > pmax) {
				pmax = p;
			}
		}
		assert (pmax >= 0) : "Can't find a reputation >= 0, aborting";
		return pmax;
	}

	public Edge getBestNode(int flowID) {
		double pmax = -1;
		Edge emax = null;
		for (Edge e : getConnections()) {
			double p = reputations.get(e).get(flowID).getValue();
			if (p > pmax) {
				pmax = p;
				emax = e;
			}
		}
		assert (pmax >= 0) : "Can't find a reputation >= 0, aborting";
		return emax;
	}

	public void addTimer(CastorMessage m) {
		CastorExpirationTimer t = new CastorExpirationTimer(this, m);
		t.startRelative(defaultTimerLength, this);
	}
}
