package projects.common.messages;

import java.util.Map;
import java.util.TreeMap;

import projects.common.MessageLogger;
import projects.common.nodes.BasicNode;
import projects.common.utils.Tools;
import sinalgo.nodes.edges.Edge;
import sinalgo.nodes.messages.Message;
import sinalgo.runtime.Global;

/**
 * Base message class to be used with BasicNodes
 * 
 * @author rjamet
 * 
 */
public class BasicMessage extends Message implements Comparable<BasicMessage> {

	/**
	 * Status of a message, either dead, alive, not done, ...
	 * 
	 * @author rjamet
	 * 
	 */
	public static enum Status {
		/**
		 * Message is still in transit
		 */
		NOTDONEYET,
		/**
		 * Message received by its intended destination
		 */
		OK,
		/**
		 * Message dropped because of too much elapsed time/hops
		 */
		TTL,
		/**
		 * Message arrived, but too late
		 */
		FORGOT,
		/**
		 * Message was eaten by a hostile node
		 */
		INTERCEPTED,
		/**
		 * Whatever happened to it, don't store/log that message in the
		 * simulation results
		 */
		DONOTLOG;
		@Override
		public String toString() {
			if (this == OK) {
				return "1";
			} else if (this == TTL) {
				return "2";
			} else if (this == FORGOT) {
				return "3";
			} else if (this == INTERCEPTED) {
				return "0";
			} else /* if (this==DONOTLOG) */{
				return "-1";
			}
		}
	};

	public static long Total_nbhops = 0;
	
	/**
	 * Who emitted that node first
	 */
	public BasicNode sender;
	
	/**
	 * Last hop (set by BasicNode.hopto()) 
	 */
	public BasicNode lastNode;

	public int hops;
	
	public static Map<Integer, Integer> duplicatas = new TreeMap<Integer, Integer>();
	public static Map<Integer, Boolean> one_delivered = new TreeMap<Integer, Boolean>();
	
	
	
	public double timeBegin;
	public double timeEnd;
	public int idMessage;
	public BasicNode destination;
	public Status st;
	

	public BasicMessage(BasicNode sender, BasicNode lastNode, int hops,
			double timeBegin, int idMessage, BasicNode dest) {
		this.sender = sender;
		this.lastNode = lastNode;
		this.hops = hops;
		this.timeBegin = timeBegin;
		this.idMessage = idMessage;
		this.destination = dest;
		st = Status.NOTDONEYET;
	}

	@Override
	public Message clone() {
		return new BasicMessage(sender, lastNode, hops, timeBegin, idMessage,
				destination);
	}

	public void hopto(Edge next) {
		assert (next.startNode instanceof BasicNode) : "Need only BasicNodes in the graph";
		hopto((BasicNode)next.startNode, (BasicNode)next.endNode);
	}

	public void hopto(BasicNode from, BasicNode to) {
		//to is useless for now
		hops++; Total_nbhops++;
		this.lastNode = from;
	}

	public void markAsDone(Status s) {
		this.st = s;
		this.timeEnd = Global.currentTime;
		MessageLogger.log(this);//may be elsewhere
	}

	public String toJSON() {
		String st = "";

		st += ("{");
		st += ("\"id\":" + this.idMessage + ",");
		st += ("\"type\":\"basic\",");
		st += ("\"h\":" + this.hops + ",");
		st += ("\"tb\":" + Tools.tojson(this.timeBegin) + ",");
		st += ("\"te\":" + Tools.tojson(this.timeEnd) + ",");
		st += ("\"f\":" + this.sender.ID + ",");
		st += ("\"d\":" + this.st + ",");
		st += ("\"o\":" + Tools.tojson(this.hops / (double)this.sender.distanceToSink) + ",");

		if (destination != null) {
			st += ("\"t\":" + this.destination.ID);
		} else {
			st += ("\"t\":-1");
		}

		st += ("}");
		return st;
	}

	@Override
	public int compareTo(BasicMessage o) {
		if (this.idMessage < o.idMessage) return -1;
		else if(this.idMessage == o.idMessage) return 0;
		else return 1;
	}

}
