package projects.common.nodes;

import projects.CASTOR.messages.CastorAckMessage;
import projects.CASTOR.messages.CastorMessage;
import projects.GBR.messages.GBRMessage;
import projects.TLCNS.messages.AckMessage;
import projects.common.timers.*;
import projects.common.utils.ConfigHelper;
import projects.common.RoutingBehaviours.RoutingBehaviour;
import projects.common.messages.*;

import java.awt.Color;
import java.util.*;

import sinalgo.configuration.WrongConfigurationException;
import sinalgo.nodes.Node;
import sinalgo.nodes.Position;
import sinalgo.nodes.edges.Edge;
import sinalgo.nodes.messages.Inbox;
import sinalgo.nodes.messages.Message;

/**
 * BasicNode : A basic abstract node, which sets all the timers and DataSeries I
 * use
 * 
 * Each non-sink node has a SenderTimer, which generates 1000 messages, one each
 * 25 steps
 * 
 * The sink (ID=1) has an EnderTimer
 * 
 * May be useful to cut sinks from that, we'll see
 */
public abstract class BasicNode extends Node {

	public static BasicNode nobody = new BasicNode() {
	};

	public int distanceToSink = Integer.MAX_VALUE;
	public static int maxDistanceToSink = -1;

	protected RoutingBehaviour rb;

	private static boolean initDone = false;

	public static int Ttl; // for now
	/**
	 * In case someone needs that
	 */
	public static Node sink;
	public static Position sinkPos;
	public static Vector<BasicNode> list = new Vector<BasicNode>(0);

	/**
	 * Stores the number of messages that have been delivered
	 */
	public double delivered_count = 0;
	/**
	 * Stores the number of messages that have been sent
	 */
	public double sent_count = 0;

	/**
	 * Add one when a message is created, remove one when one is destroyed Used
	 * by EnderTimer to check if the simulation is done
	 */
	public static int rolling_messages = 0;

	/**
	 * Variable extracted from the config.xml file, makes nodes avoid directly
	 * sending messages to their last senders when there's a choice
	 */
	public static boolean noBacktrack = true;

	/**
	 * Instead of 1/nbneighbors, the probability of choosing a node n in a
	 * "random" routing is :
	 * 
	 * <code>
	 * 				(1/sqrt(degree(n)))
	 * ---------------------------------------
	 * sum for(x:neighbors){1/sqrt(degree(x))}
	 * </code>
	 */
	public static boolean useLD = false;

	public Collection<Edge> blacklist = new ArrayList<Edge>();

	/**
	 * May be changed, may be multiple or no sinks, ...
	 * 
	 * @return true if the node is a sink.
	 */
	public boolean isSink() {
		return false; // for now
	}

	/**
	 * Indicates whether this node sends messages or not may be overloaded if
	 * you do time-based behaviour changing
	 * 
	 * @return Whether this node is a source right now
	 */
	public boolean sendsMessages() {
		return true;
	}

	@Override
	/** Wraps BasicNode.send to monitor how many messages are currently being in transit*/
	public void send(Message m, Node target) {
		rolling_messages++;
		super.send(m, target);
	}

	/**
	 * Returns a collection of the edges of the node, instead of that terrible
	 * Connections thing
	 * 
	 * @return All the outgoing edges of the node
	 */
	public Collection<Edge> getConnections() {
		Vector<Edge> edges = new Vector<Edge>();
		Iterator<Edge> it = this.outgoingConnections.iterator();
		while (it.hasNext()) {
			Edge edge = it.next();
			if (!blacklist.contains(edge)) {
				edges.add(edge);
			}
		}
		return edges;
	}

	/**
	 * Removes all edges leading to the specified node from the collection, and
	 * returns it for convenience (but it's still modified).
	 * 
	 * @param v
	 *            The collection to modify
	 * @param from
	 *            The node that should be removed
	 */
	public static Collection<? extends Edge> noBacktrackCheck(
			Collection<? extends Edge> v, Node from) {
		assert (from != null) : "Should not happen";
		Iterator<? extends Edge> it = v.iterator();
		Edge e;
		while (it.hasNext()) {
			e = it.next();
			if (v.size() > 1 && e.endNode.equals(from)) {
				it.remove();
			}
		}
		return v;
	}

	/**
	 * Chooses a random edge to send a message to from the collection v,
	 * according to the local degree politics of the current run. Should already
	 * have made backtracking choices by then.
	 * 
	 * @return The chosen edge
	 */
	public Edge chooseRandomFromCollection(Collection<? extends Edge> v) {

		double sum = 0;
		double sumdegrees = 0;

		if (useLD) { // prepare the divisor
			for (Edge e : v)
				sumdegrees += 1.0 / Math.sqrt(e.endNode.outgoingConnections
						.size());
		}

		if (useLD) {
			for (Edge e : v) { // get the total
				sum += (1.0 / Math.sqrt(e.endNode.outgoingConnections.size()))
						/ sumdegrees;
			}
		} else {
			sum = v.size();
		}

		double rnd = sinalgo.tools.statistics.UniformDistribution.nextUniform(
				0, sum);

		for (Edge e : v) { // and find the random one
			if (useLD) {
				rnd -= (1.0 / Math.sqrt(e.endNode.outgoingConnections.size()))
						/ sumdegrees;
			} else {
				rnd -= 1;
			}
			if (rnd <= 0) {
				return e;
			}
		}

		// should never happen
		throw new RuntimeException("Programming error above, sum=" + sum
				+ " rand=" + rnd);
	}

	/**
	 * Chooses a random edge to send a message to, according to the
	 * backtracking/local degree politics of the current run.
	 * 
	 * @see ChooseRandomFromCollection
	 */
	public Edge chooseRandom(Node from, Collection<? extends Edge> v) {
		if (this.outgoingConnections.size() == 0) {
			throw new RuntimeException(
					"Neighborless node : graph isn't connected");
		} else {
			if (from != null && noBacktrack && v.size() > 1) {
				return chooseRandomFromCollection(noBacktrackCheck(v, from));
			} else {
				return chooseRandomFromCollection(v);
			}
		}
	}

	@Override
	public void handleMessages(Inbox inbox) {
		while (inbox.hasNext()) {
			Message msg_receive = inbox.next();
			this.action(msg_receive);
		}
	}

	/**
	 * Action : called once per message, does basic treatment such as eating the
	 * message if the node is a sink, or route to the next hop otherwise. If you
	 * overload it, you should still call the super method.
	 */
	protected void action(Message msg_receive) {
		assert (msg_receive instanceof BasicMessage) : "msg_receive isn't a supported type of message";
		rolling_messages--;
		rb.act((BasicMessage) msg_receive);

	}

	/**
	 * Function called at each consumption of a message, doing the logging,
	 * output, and message number tracking (for termination). If you overload
	 * it, make sure you still call the superfunction.
	 */
	public void eatMessage(BasicMessage msg, BasicMessage.Status s) {

//		System.out.println("Node "+ID+" eats m "+msg.idMessage +" status "+s);

		
		if (msg instanceof BasicMessage) {
			BasicMessage bm = (BasicMessage) msg;
//gros changements
			if (bm instanceof CastorMessage || bm instanceof CastorAckMessage) {
				BasicMessage.duplicatas.put(bm.idMessage,BasicMessage.duplicatas.get(bm.idMessage)-1);
				if (s == BasicMessage.Status.DONOTLOG) {
					//nothing
				} else  if ( s == BasicMessage.Status.OK) {
					if (!BasicMessage.one_delivered.get(msg.idMessage)) {
						msg.markAsDone(s); 
					}
					BasicMessage.one_delivered.put(msg.idMessage,true);
				} else if (s == BasicMessage.Status.TTL || s == BasicMessage.Status.FORGOT || s == BasicMessage.Status.INTERCEPTED) {
					if (BasicMessage.duplicatas.get(msg.idMessage) == 0) {
						if (!BasicMessage.one_delivered.get(msg.idMessage)) {
							msg.markAsDone(s); 
						} else {
							//nothing
						}
					} else {
						//nothing
					}
				}
			} else {
			
				if (s != BasicMessage.Status.DONOTLOG) {
					msg.markAsDone(s);
				}
			}

			if (!(msg instanceof AckMessage ||msg instanceof CastorAckMessage || msg instanceof GBRMessage)) {
				if (s == BasicMessage.Status.OK) {
					assert(msg.sender != null);
					msg.sender.delivered_count++;
				}
			}
		}

	}

	/**
	 * Initialize the node, start timers, and parse the configuration (hackish)
	 * TODO: are neighborhoods done yet ?
	 */
	@Override
	public void init() {
		if (!initDone) {
			initDone = true;
			(new EnderTimer()).startGlobalTimer(1);
			ConfigHelper.parseConfig();
		}
		if (isSink()) {
			sink = this;
			sinkPos = this.getPosition();
		}
		list.setSize(ID + 1);
		list.add(ID, this);
		startTimers();
	}

	protected void startTimers() {
		if (!isSink()) {
			(new SenderTimer(this)).startRelative(1, this);
		}
	}

	@Override
	public String toString() {
		String s = "Node(" + this.ID + "), d=" + this.distanceToSink + ", [";
		Iterator<Edge> edgeIter = this.getConnections().iterator();
		while (edgeIter.hasNext()) {
			Edge e = edgeIter.next();
			Node n = e.endNode;
			s += n.ID + " ";
		}
		edgeIter = this.blacklist.iterator();
		s += "] BL:[";
		while (edgeIter.hasNext()) {
			Edge e = edgeIter.next();
			Node n = e.endNode;
			s += n.ID + " ";
		}
		return s + "]";
	}

	@Override
	public void checkRequirements() throws WrongConfigurationException {
	}

	/**
	 * Returns the color of the node, for drawing purposes.
	 */
	@Override
	public Color getColor() {
		if (isSink()) { // sink
			return Color.BLUE;
		} else { // not a sink
			return new Color(0, (float) 0.5, 0);
		}
	}

	public void atTheEnd() {
		// Unused for now, called only once for each sink
	}

	@Override
	/**
	 * Called after changes in the network or the initialization. 
	 */
	public void neighborhoodChange() { 
		rb.init();
		if (this.isSink()) {
			
			@SuppressWarnings("unchecked")
			Vector<BasicNode> nodesNotSeen = (Vector<BasicNode>) list.clone();
			Stack<BasicNode> todo = new Stack<BasicNode>();

			this.distanceToSink = 0;
			todo.push(this);
			nodesNotSeen.remove(this);
			
			while(!todo.isEmpty()) {
				BasicNode currentNode = todo.remove(0);
				nodesNotSeen.remove(currentNode);
				
				for (Edge e : currentNode.getConnections()) {
					BasicNode n = ((BasicNode) e.endNode);
					if (nodesNotSeen.contains(n)) {
						nodesNotSeen.remove(n);
						todo.push(n);
						n.distanceToSink = currentNode.distanceToSink + 1;
						maxDistanceToSink = n.distanceToSink;
					}
				}
			}
			
		}
	}

	@Override
	public void postStep() {
		// try to keep it empty, asynchronous mode is faster and doesn't run
		// this
	}

	@Override
	public void preStep() {
		// try to keep it empty, asynchronous mode is faster and doesn't run
		// this
	}

	/**
	 * What the node will output when asked (ends up in node_data.json)
	 */
	public void JSONOutput() {
		assert (sent_count >= delivered_count) : "More delivered messages than sent messages: s"+sent_count+" r"+delivered_count;
		if (sent_count == 0) {
			sent_count++;
		}
		System.out.println("{\"t\":\""
				+ JSONName()
				+ "\","
				+ " \"dr\":"
				+ projects.common.utils.Tools
						.tojson((delivered_count / (sent_count))) + ","
				+ "\"id\":" + ID + "," + " \"d\":"
				+ this.outgoingConnections.size()
				/*
				 * + " \"blsize\":" + this.blacklist.size() + "," +
				 * " \"postbld\":" + this.getConnections().size()
				 */+ "," + this.JSONNeighbors() + "},");
	}

	/**
	 * Returns a JSONed list of neighbors id, not culling on blacklists example
	 * : "[1,2,3,5]"
	 */
	public String JSONNeighbors() {
		String r = "";
		r += "\"neighbors\":[";
		Iterator<Edge> i = this.outgoingConnections.iterator();
		while (i.hasNext()) {
			Edge e = i.next();
			r += (e.endNode.ID);
			if (i.hasNext())
				r += ",";
		}
		r += ("]");
		return r;
	}

	public String JSONName() {
		return "basic";
	}

}
