package projects.common.nodes;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

import projects.common.messages.BasicMessage;
import projects.common.utils.TabuList;
import sinalgo.nodes.edges.Edge;

/**
 * 
 * BasicTabuNode is the base class for every node that uses a counter and/or a
 * tabulist per edge.
 * 
 * @author rjamet
 * 
 */
public abstract class BasicTabuNode extends BasicNode {

	/**
	 * A hashmap mapping a counter (double) for each outgoing edge. This counter
	 * is then used in routing, see select().
	 */
	public HashMap<Edge, Double> counters;

	/**
	 * A hashmap mapping a TabuList for each outgoing edge.
	 */
	public HashMap<Edge, TabuList> tl;

	/**
	 * Changes the probability of choosing the next node
	 */
	public static double alpha = 1;

	/**
	 * Variable set after the settings in the config file. Initial counter
	 * value.
	 */
	public static double CounterInit = 1;

	@Override
	public void init() {
		counters = new HashMap<Edge, Double>();
		tl = new HashMap<Edge, TabuList>();
		super.init();
	}

	/**
	 * Initializes the TL/Counter for this edge if needed
	 * 
	 * @param e
	 */
	public void initCheck(Edge e) {
		if (tl.get(e) == null || counters.get(e) == null) {
			tl.put(e, TabuList.build());
			counters.put(e, CounterInit/this.getConnections().size());
		}
	}

	/**
	 * 
	 * Selects the next edge that this message will take. Does so based on the
	 * counters : probabilities of taking an edge are proportionnal to the
	 * counter of that edge, after adjustements for the local degrees and
	 * backtracking politics.
	 * 
	 * May also contain the inverse of counters, depending on the algorithm. Check the update policy
	 * 
	 * @param m
	 *            The message that needs to be routed
	 * @return The destination edge, or null if error.
	 */
	public Edge select(BasicMessage m) {

		double sum = 0;
		double sumdegrees = 0;
		Collection<Edge> v = getConnections();
		/*if (m.lastNode != null && noBacktrack && v.size() > 1) {
			noBacktrackCheck(v, m.lastNode);
		}*/

		for (Edge e : v) {
			initCheck(e);
		}

		if (useLD) {
			for (Edge e : v) {
				sumdegrees += Math.pow(
						counters.get(e)
								/ Math.sqrt(e.endNode.outgoingConnections
										.size()), alpha);
			}
		}

		for (Edge e : v) {
			if (useLD) {
				sum += Math.pow(
						counters.get(e)
								/ Math.sqrt(e.endNode.outgoingConnections
										.size()), alpha)
						/ sumdegrees;
			} else {
				sum += Math.pow(counters.get(e), alpha);
			}
		}

		double rnd = sinalgo.tools.statistics.UniformDistribution.nextUniform(
				0, sum);
		assert (!Double.isNaN(rnd)) : "NaN generated by RNG : this is bad (infinite counter ?)";

		for (Edge e : v) {
			if (useLD) {
				rnd -= (Math.pow(
						counters.get(e)
								/ Math.sqrt(e.endNode.outgoingConnections
										.size()), alpha))
						/ sumdegrees;
			} else {
				rnd -= Math.pow(counters.get(e), alpha);
			}

			if (rnd <= 0)
				return e;
		}
		assert (false) : "No results for select (neighborless node ?)";
		return null; // in case of no backtrack and one neighbor
						// only
	}

	/**
	 * Returns a string containing a JSON object for the node, which explicits
	 * all the connexions and counters associated to them.
	 */

	@Override
	public void JSONOutput() {

		assert (sent_count >= delivered_count) : "More delivered than sent messages";

		if (sent_count == 0) {
			sent_count++; // division par 0
		}

		System.out.println("{\"id\":"
				+ ID
				+ ", \"t\":\" "+ JSONName() +"\", \"d\":"
				+ this.outgoingConnections.size()
				+ ","
				+ " \"dr\":"
				+ projects.common.utils.Tools
						.tojson((delivered_count / (sent_count))) + ","
				/* + JSONCounters() + "," */ + JSONNeighbors() + "},");
	}

	public String JSONCounters() {
		
		String r = "";
		r += "\"counters\":{";
		Iterator<Edge> i = this.outgoingConnections.iterator();
		while (i.hasNext()) {
			Edge e = i.next();
			initCheck(e);
			r += ("\"" + e.endNode.ID + "\":");
			r += (Math.max(Math.round(counters.get(e)),
					Math.round(1 / counters.get(e))));
				r += (",");
		}
		r += ("}");
		return r; 
	}

	@Override
	public String JSONName() {
		return "tabu";
	}

}
