package projects.defaultProject.models.messageTransmissionModels;

import java.util.*;
import sinalgo.configuration.CorruptConfigurationEntryException;
import sinalgo.nodes.Node;
import sinalgo.nodes.messages.Message;
import sinalgo.runtime.Global;
import sinalgo.tools.statistics.Distribution;

/**
 * transmission model whose delivery time is defined through a distribution.
 * Fifo transmission channels is preserved.
 * <p>
 * This class expects an entry in the configuration file that describes the
 * distribution of the delivery time.
 * 
 * <pre>
 * 		&lt;FifoRandom distribution="Uniform" min="1" max="10"/&gt;
 * </pre>
 * 
 * a random time that follows the above distribution is added to the last time a
 * message was sent through this channel.
 */
public class FifoRandom extends sinalgo.models.MessageTransmissionModel {

	public final static int nodeTupleCode(Node from, Node to) {
		return (from.ID << 15) ^ to.ID;
		//unique as long as 0 <= ID < 32k (on est larges)
	}
	
	private static Map<Integer, Double> lastSent; 
	Distribution dist;

	/**
	 * Creates a new FifoRandom transmission model instance and reads the config
	 * for this object from the config file.
	 * 
	 * @throws CorruptConfigurationEntryException
	 */
	public FifoRandom() throws CorruptConfigurationEntryException {
		dist = Distribution.getDistributionFromConfigFile("FifoRandom");
		lastSent = new HashMap<Integer, Double>();// plus rapide
	}

	@Override
	public double timeToReach(Node startNode, Node endNode, Message msg) {
		int edge = nodeTupleCode(startNode, endNode);
		Double t = lastSent.get(edge);

		if (t == null || t <= Global.currentTime) {
			t = Global.currentTime;
		}
		
		t =  t + dist.nextSample();
		lastSent.put(edge, t);
		double delay = t - Global.currentTime;
		//assert(delay < 500):"Strong congestion : more than 500 delay on a link.";
		return delay;
	}
}